/*
 * This file belongs to the XQuark distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquark.org.
 */

package org.xquark.bridge;

import java.io.*;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLClassLoader;
import java.sql.Connection;
import java.sql.Driver;
import java.sql.SQLException;
import java.util.*;

import javax.xml.parsers.ParserConfigurationException;
import javax.xml.parsers.SAXParserFactory;

import org.xml.sax.*;
import org.xquark.extractor.Extractor;
import org.xquark.extractor.ExtractorReadOnlyConnection;
import org.xquark.extractor.common.Log;
import org.xquark.extractor.common.Trace;
import org.xquark.serialize.XMLSerializer;
import org.xquark.util.misc.Debug;
import org.xquark.xml.xdbc.*;

/**
 * Objects of this class encapsulate a JDBC connection and provide access to both
 * the XQuery and insertion capabilities of the XQBridge.<BR>
 * Objects of this class are thread-safe and can be reused multiple times.
 * However, as they encapsulate a JDBC connection, it is recommended to use
 * different objects with different JDBC connections for different threads.
 */
public class XQBridge {
    private static final String RCSRevision = "$Revision: 1.9 $";
    private static final String RCSName = "$Name:  $";

    private static final String HELP_OPT1 = "-help";
    private static final String HELP_OPT2 = "-?";
    private static final String CONF_OPT = "-conf";
    private static final String PROP_OPT1 = "-p";
    private static final String PROP_OPT2 = "-properties";
    private static final String DRIVER_OPT = "-driver";
    private static final String JDBC_OPT = "-db";
    private static final String JDBC_OPT2 = "-url";
    private static final String USER_OPT = "-user";
    private static final String PASSWORD_OPT = "-password";
    private static final String QUERY_OPT1 = "-query";
    private static final String QUERY_OPT2 = "-q";
    private static final String INSERT_OPT1 = "-insert";
    private static final String INSERT_OPT2 = "-i";
    private static final String FILE_OPT1 = "-file";
    private static final String FILE_OPT2 = "-f";
    private static final String DEBUG_OPT = "-debug";
	private static final String CLASSPATH_OPT1 = "-cp";
	private static final String CLASSPATH_OPT2 = "-classpath";
	
	/** List of supported JDBC drivers */
	
	private static final String[] DRIVER_LIST = {
		"com.mysql.jdbc.Driver",
		"com.microsoft.jdbc.sqlserver.SQLServerDriver",
		"com.sybase.jdbc2.jdbc.SybDriver",
		"oracle.jdbc.driver.OracleDriver"
	};

    private static SAXParserFactory parserFactory;

    static {
        parserFactory = SAXParserFactory.newInstance();
        parserFactory.setNamespaceAware(true);
    }

    private Connection jdbcConnection;
    private String configurationFile;
    private XMLConnection xmlConnection = null;
    private PrintStream log = null;
    private ClassLoader classLoader = null;

    /* The constructor for the XQBridge object.
     * @param jdbcConn the JDBC connection used to access the database.
     */
    public XQBridge(Connection jdbcConn) {
        if (jdbcConn == null)
            throw new NullPointerException("Illegal null value for connection");
        jdbcConnection = jdbcConn;
    }

    /* The constructor for the XQBridge object. This constructor allocates its own JDBC connection.
     * The close() method must be called on the object to free the JDBC resources.
     * @param confFile the XQBridge configuration file
     */
    public XQBridge(String confFile) {
        if (confFile == null)
            throw new NullPointerException("Illegal null value for configuration file");
        configurationFile = confFile;
    }

    /* The constructor for the XQBridge object. The specified JDBC connection is used to access the database,
     * and overrides connection information provided in the configuration file.
     * @param jdbcConn the JDBC connection used to access the database.
     * @param confFile the XQBridge configuration file
     */
    public XQBridge(Connection jdbcConn, String confFile) {
        if (jdbcConn == null)
            throw new NullPointerException("Illegal null value for connection");
        if (confFile == null)
            throw new NullPointerException("Illegal null value for configuration file");
        jdbcConnection = jdbcConn;
        configurationFile = confFile;
    }

    /**
     * Returns the JDBC connection used to access the underlying database. 
     * Returns null when no connection to the database is available.
     * @return a JDBC connection
     */
    public Connection getConnection() {
        if (jdbcConnection == null) {
            try {
                ExtractorReadOnlyConnection extractorConn =
                    (ExtractorReadOnlyConnection) getXMLConnection();
                jdbcConnection = extractorConn.getJdbcConnection();
            }
            catch (XMLDBCException ex) {
                // Empty on purpose
            }
        }
        return jdbcConnection;
    }

    /**
     * Returns the XMLDBC connection used to execute XQuery statements over the database.
     * @return an XMLConnection object
     */
    public XMLConnection getXMLConnection() throws XMLDBCException {
        if (xmlConnection == null) {
            xmlConnection =
                new Extractor(jdbcConnection, configurationFile, classLoader).getConnection();
        }
        return xmlConnection;
    }

    public void close() throws XMLDBCException {
        if (xmlConnection != null && !xmlConnection.isClosed())
            xmlConnection.close();
    }
    
    /**
     * Set the class loader used for loading drivers, mappings and user generators.
     * @param loader ClassLoader 
     */
    public void setClassLoader(ClassLoader loader) {
    	classLoader = loader;
    }
    
    /**
     * Returns the class loader used for loading drivers, mappings and user generators.
     */
    public ClassLoader getClassLoader() {
    	return classLoader != null ? classLoader : this.getClass().getClassLoader();
    }

    /**
     * Creates a mapping object containing the processed version of the specified mapping file.
     * This object can subsequently be used to store XML documents in the database, according
     * to the specified mapping.<BR>
     * The mapping file is specified as a SAX input source. If the systemId property of the input source is set,
     * the object will first be looked up, based on this property, in the mapping cache, and only be created if
     * not found.
     * @deprecated Prefer the string methods because they do not pose a problem
     * for resolving relative URIs in mapping file schemaLocation attribute.
     * @param source a SAX input source encapsulating the mapping file.
     * @return a Mapping object
     * @throws IOException if an IO error occured while accessing the mapping file
     * @throws SAXException if a SAX error occured while parsing the mapping file
     * @throws SQLException if an SQL error occured while accessing the database
     */
    public Mapping getMapping(InputSource source)
        throws SAXException, SQLException, IOException {
        return getMapping(source, false);
    }

    /**
     * Creates a mapping object containing the processed version of the specified mapping file.
     * This object can subsequently be used to store XML documents in the database, according
     * to the specified mapping.<BR>
     * The mapping file is specified as a SAX input source. If the systemId property of the input source is set,
     * and the reload flag is false, the object will first be looked up, based on this property,
     * in the mapping cache, and only be created if not found.
     * @deprecated Prefer the string methods because they do not pose a problem
     * for resolving relative URIs in mapping file schemaLocation attribute.
     * @param source a SAX input source encapsulating the mapping file.
     * @param reload true if the mapping file must be reloaded even if found in the cache.
     * @return a Mapping object
     */
    public Mapping getMapping(InputSource source, boolean reload)
        throws SAXException, SQLException, IOException {
        return new Mapping(getConnection(), source, reload, getClassLoader());
    }

    /**
     * Creates a mapping object containing the processed version of the specified mapping file.
     * This object can subsequently be used to store XML documents in the database, according
     * to the specified mapping.<BR>
     * the object will first be looked up, based on the file URI, in the mapping cache,
     * and only be created if not found.
     * @param fileURI the URI of the mapping file.
     * @return a Mapping object
     * @throws IOException if an IO error occured while accessing the mapping file
     * @throws SAXException if a SAX error occured while parsing the mapping file
     * @throws SQLException if an SQL error occured while accessing the database
     */
    public Mapping getMapping(String fileURI)
        throws SAXException, SQLException, IOException {
        return getMapping(fileURI, false);
    }

    /**
     * Creates a mapping object containing the processed version of the specified mapping file.
     * This object can subsequently be used to store XML documents in the database, according
     * to the specified mapping.<BR>
     * If the reload flag is false, the object will first be looked up, based on the file URI,
     * in the mapping cache, and only be created if not found.
     * @param fileURI the URI of the mapping file.
     * @param reload true if the mapping file must be reloaded even if found in the cache.
     * @return a Mapping object
     */
    public Mapping getMapping(String fileURI, boolean reload)
        throws SAXException, SQLException, IOException {
        return new Mapping(jdbcConnection, fileURI, reload, getClassLoader());
    }

    /* This part of the file deals with command-line execution */

    /**
     * The method called from the command line to run the XQBridge as a standalone program.
     * Use the -help option to display a list of available options.
     * @param args the command line parameters
     */
    public static void main(String[] args) {
        String propFile = "XQBridge.properties";
        boolean defaultPropFile = true;
        String configFile = null;
        String batchFile = null;
        String driver = null;
        String url = null;
        String user = null;
        String password = null;
        String classpath = null;
        ClassLoader classloader = XQBridge.class.getClassLoader();
        boolean isQuery = false;
        boolean isInsert = false;
        ArrayList parameters = new ArrayList();

        for (int i = 0; i < args.length; i++) {
            if (!args[i].startsWith("-")) {
                // not an option
                parameters.add(args[i]);
            }
            else {
                // Check options are in first position
                if (parameters.size() > 0) {
                    usage();
                    System.err.println(
                        "Options must appear before the command parameters.");
                    System.exit(-1);
                }
                // one-parameter option
                if (HELP_OPT1.equals(args[i]) || HELP_OPT2.equals(args[i])) {
                    usage();
                    System.exit(0);
                }
                else if (
                    QUERY_OPT1.equals(args[i]) || QUERY_OPT2.equals(args[i])) {
                    isQuery = true;
                }
                else if (
                    INSERT_OPT1.equals(args[i])
                        || INSERT_OPT2.equals(args[i])) {
                    isInsert = true;
                }
                else {
                    // two-parameter option
                    if (i + 1 == args.length) {
                        System.err.println(
                            "Missing value for option " + args[i]);
                        usage();
                        System.exit(-1);
                    }
                    if (CONF_OPT.equals(args[i])) {
                        configFile = args[++i];
                    }
                    else if (
                        PROP_OPT1.equals(args[i])
                            || PROP_OPT2.equals(args[i])) {
                        propFile = args[++i];
                        defaultPropFile = false;
                    }
                    else if (DEBUG_OPT.equals(args[i])) {
                        Debug.ON = true;
                        Log.enable();
                        Trace.traceClass("pippo");
                    }
                    else if (
                        FILE_OPT1.equals(args[i])
                            || FILE_OPT2.equals(args[i])) {
                        batchFile = args[++i];
                    }
                    else if (DRIVER_OPT.equals(args[i])) {
                        driver = args[++i];
                    }
                    else if (
                        JDBC_OPT.equals(args[i])
                            || JDBC_OPT2.equals(args[i])) {
                        url = args[++i];
                    }
                    else if (USER_OPT.equals(args[i])) {
                        user = args[++i];
                    }
                    else if (PASSWORD_OPT.equals(args[i])) {
                        password = args[++i];
                    }
					else if (
						CLASSPATH_OPT1.equals(args[i])
							|| CLASSPATH_OPT2.equals(args[i])) {
						classpath = args[++i];
					}
                    else {
                        // Unknown option
                        System.err.println("Invalid option " + args[i]);
                        usage();
                        System.exit(-1);
                    }
                }
            }
        }
        if (isQuery == isInsert) {
            System.err.println(
                "One and only one of -q(uery) and -i(nsert) must be specified");
            System.exit(-1);
        }

        Properties props = new Properties();
        try {
            props.load(new FileInputStream(propFile));
        }
        catch (IOException e) {
            InputStream inStream = null;
            try {
                inStream = XQBridge.class.getResourceAsStream("/" + propFile);
                if (inStream != null)
                    props.load(inStream);
            }
            catch (IOException ex) {
                inStream = null;
            }
            if (configFile == null
                && inStream == null
                && (driver == null
                    || url == null
                    || user == null
                    || password == null)) {
                System.err.println(
                    "error : configuration file "
                        + propFile
                        + " couldn't be found.");
                System.err.println(
                    "(neither using file name nor in CLASSPATH root)");
                System.exit(-1);
            }
        }

        if (configFile == null || !defaultPropFile) {
            if (driver == null)
                driver = props.getProperty("database.driver");
            if (url == null)
                url = props.getProperty("database.url");
            if (user == null)
                user = props.getProperty("database.user");
            if (password == null)
                password = props.getProperty("database.password");
        }
        
        if (classpath == null) 
        	classpath = props.getProperty("xqbridge.classpath");
        	
        if (classpath != null) {
        	classloader = new URLClassLoader(parseClassPath(classpath), classloader);
        }

		ArrayList drivers = new ArrayList();
		for (int i = 0; i < DRIVER_LIST.length; i++) {
			try {
				Class c = Class.forName(DRIVER_LIST[i], true, classloader);
				drivers.add(c.newInstance());
			} catch (ClassNotFoundException e) {
            } catch (InstantiationException e) {
            } catch (IllegalAccessException e) {
            }
		}

		Connection jdbc = null;
		if (url != null && user != null && password != null) {
			Properties jdbcProps = new Properties();
			jdbcProps.put("user", user);
			jdbcProps.put("password", password);
			try {
				for (int i = 0; jdbc == null && i < drivers.size(); i++) {
					Driver d = (Driver)drivers.get(i);
					if (d.acceptsURL(url)) {
						jdbc = d.connect(url, jdbcProps);
					}
				}
			}
			catch (SQLException e) {
				jdbc = null;
			}
		}
        if (configFile != null) {
            try {
                configFile = new File(configFile).toURL().toString();
            }
            catch (MalformedURLException ex) {
                System.err.println("Invalid configuration file specification.");
                System.exit(-1);
            }
        }
        XQBridge bridge = null;
        if (jdbc == null) {
            if (configFile == null) {
                System.err.println(
                    "Could not connect to database "
                        + url
                        + ". Check your JDBC parameters.");
                System.exit(-1);
            }
            bridge = new XQBridge(configFile);
        }
        else {
            if (configFile == null) {
                bridge = new XQBridge(jdbc);
            }
            else {
                bridge = new XQBridge(jdbc, configFile);
            }
        }
        
		bridge.setClassLoader(classloader);
        
        List parametersList = null;
        if (batchFile != null) {
            parametersList = parseBatchFile(batchFile, isQuery);
        }
        else {
            checkParameters(parameters, isQuery);
            parametersList = new ArrayList(1);
            parametersList.add(parameters);
        }
        int result = 0;
        if (isQuery) {
	        result = bridge.executeQuery(parametersList);
        }
        else {
        	result = bridge.executeInsert(parametersList);
        }
        System.exit(result);
    }

    private static void checkParameters(List params, boolean isQuery) {
        int size = params.size();
        if (isQuery) {
            if ((isQuery && (size < 1 || size > 2))
                || (!isQuery && size < 2)) {
                System.err.println("Wrong number of parameters.");
                System.exit(-1);
            }
        }
    }

    private static List parseBatchFile(String batchFile, boolean isQuery) {
        ArrayList result = new ArrayList();
        BufferedReader reader = null;
        try {
            File f = new File(batchFile);
            reader = new BufferedReader(new FileReader(f));
        }
        catch (IOException ex) {
            System.err.println("Could not find specified batch file");
            System.exit(-1);
        }
        try {
            while (true) {
                String line = reader.readLine();
                if (line == null)
                    break;
                if (line.startsWith("#") || line.trim().length() == 0)
                    continue;
                java.util.StringTokenizer it =
                    new java.util.StringTokenizer(line);
                ArrayList params = new ArrayList();
                while (it.hasMoreTokens()) {
                    params.add(it.nextToken());
                }
                checkParameters(params, isQuery);
                result.add(params);
            }
        }
        catch (IOException e) {
            System.err.println("Error while reading specified batch file");
            System.exit(-1);
        }
        return result;
    }

    private static String readQueryFile(String fileName) {
        try {
            StringBuffer retVal = new StringBuffer();
            String tmpStr;
            BufferedReader in = new BufferedReader(new FileReader(fileName));
            while ((tmpStr = in.readLine()) != null) {
                if (retVal.length() > 0)
                    retVal.append('\n');
                retVal.append(tmpStr);
            }
            in.close();
            return retVal.toString();
        }
        catch (FileNotFoundException ex) {
            System.err.println("Could not find query file " + fileName);
        }
        catch (IOException ex) {
            System.err.println("Error while reading query file " + fileName);
        }
        return null;
    }

    private int executeQuery(List parametersList) {
    	int result = 0;
		XMLConnection connection = null;
		XMLStatement stat = null;
		try {
			connection = getXMLConnection();
			stat = connection.createStatement();
		} catch (XMLDBCException ex) {
			System.err.println(
				"Error while creating XQuery statement");
			printError(ex);
			result = -1;
		}
        Iterator it = parametersList.iterator();
        while (it.hasNext()) {
            List params = (List) it.next();
            String queryFile = (String) params.get(0);
            String resultFile = null;
            if (params.size() > 1)
                resultFile = (String) params.get(1);
            String query = readQueryFile(queryFile);
            if (query != null) {
                OutputStream out = System.out;
                try {
                    if (resultFile != null) {
                        out =
                            new BufferedOutputStream(
                                new FileOutputStream(new File(resultFile)));
                    }
                }
                catch (IOException ex) {
                    out = null;
                    System.err.println(
                        "Could not open result file " + resultFile);
					result = -1;
                }
                if (out != null) {
                    try {
                        stat.setBaseURI(new File(queryFile).toURL().toString());
                        XMLResultSet rs = stat.executeQuery(query);
                        XMLSerializer handler = new XMLSerializer(out);
                        rs.setContentHandler(handler);
                        handler.startDocument();
                        long rtime = System.currentTimeMillis();
                        while (rs.hasNext())
                            rs.nextAsSAX();
                        handler.endDocument();
                        rs.close();
                        if (out != System.out)
                            out.close();
                    } catch (MalformedURLException e) {
						System.err.println(
							"Error while accessing query file " + queryFile);
						printError(e);
						result = -1;
                    } catch (XMLDBCException e) {
						System.err.println(
							"Error while processing query file " + queryFile);
						printError(e);
						result = -1;
                    } catch (UnsupportedEncodingException e) {
						System.err.println(
							"Illegal encoding for query result for " + queryFile);
						printError(e);
						result = -1;
                    } catch (SAXException e) {
						System.err.println(
							"Error while processing query result for " + queryFile);
						printError(e);
						result = -1;
                    } catch (IOException e) {
						System.err.println(
							"Error while closing result stream");
						printError(e);
						result = -1;
                    }
                }
            } else {
            	result = -1;
            }
        }
		try {
			if (stat != null) stat.close();
		} catch (XMLDBCException ex) {
			System.err.println(
				"Error while closing XQuery statement");
			printError(ex);
			result = -1;
		}
		return result;
    }

    private int executeInsert(List parametersList) {
        int result = 0;
        Iterator it = parametersList.iterator();
        while (it.hasNext()) {
            List params = (List) it.next();
            String mapFile = (String) params.get(0);
            Mapping mapping = null;
            try {
                mapping = getMapping(mapFile);
            } catch (SQLException ex) {
                System.err.println(
                    "Database access error while processing mapping file " + mapFile);
                printError(ex);
                result = -1;
            } catch (SAXException ex) {
				System.err.println(
					"Error while processing mapping file " + mapFile);
				printError(ex);
				result = -1;
            } catch (IOException ex) {
				System.err.println(
					"Error while accessing mapping file " + mapFile);
				printError(ex);
				result = -1;
            }
            if (mapping != null) {
				String xmlFile = null;
                try {
					Mapper mapper = mapping.getMapper();
                    XMLReader reader =
                        parserFactory.newSAXParser().getXMLReader();
                    reader.setContentHandler(mapper);
                    for (int i = 1; i < params.size(); i++) {
                        xmlFile = (String) params.get(i);
                        reader.parse(xmlFile);
                    }
					mapper.close();
					// Mapping is not closed on purpose (to use the cache mechanism)
                } catch (XMLDBCException ex) {
					System.err.println(
						"Error while accessing mapping information for " + xmlFile);
					printError(ex);
					result = -1;
                } catch (SAXException ex) {
					System.err.println(
						"Error while processing XML file " + xmlFile);
					printError(ex);
					result = -1;
                } catch (ParserConfigurationException ex) {
					System.err.println(
						"Cannot create XML parser for " + xmlFile);
					printError(ex);
					result = -1;
                } catch (IOException ex) {
					System.err.println(
						"Error while accessing XML file " + xmlFile);
					printError(ex);
					result = -1;
                }
            }
        }
        return result;
    }

    /** Print an error message to the standard error output.
     */
    private static void printError(Throwable e) {
        if (e != null) {
        	if (e.getMessage() != null) {
				/* print exception message */
				System.err.println("Exception: "+e.getMessage());
        	}

            /* printing location info */
            if (e instanceof SAXParseException)
                System.err.println(
                    "Element position in XML source is : Line "
                        + ((SAXParseException) e).getLineNumber()
                        + ", Column "
                        + ((SAXParseException) e).getColumnNumber()
                        + ".");
//            /* underlying exceptions */
//            if (e instanceof XMLDBCException)
//                printError(((XMLDBCException) e).getException());
//            else if (e instanceof SAXException)
//                printError(((SAXException) e).getException());
        }
    }
    
    private static URL[] parseClassPath(String classpath) {
		String pathSeparator = System.getProperty("path.separator");
		ArrayList urls = new ArrayList();
		if (classpath != null) {
			StringTokenizer tokenizer = new StringTokenizer(classpath, pathSeparator);
			while (tokenizer.hasMoreTokens()) {
				File file = new File(tokenizer.nextToken());
				try {
					urls.add(file.getCanonicalFile().toURL());
				} catch (MalformedURLException e) {
                } catch (IOException e) {
                }
			}
		}
		return (URL[]) urls.toArray(new URL[0]);
    }

    private static void usage() {
        System.out.println(
            "Usage: java org.xquark.bridge.XQBridge -q(uery) <common options> <file1> [<file2>]");
        System.out.println(
            "       java org.xquark.bridge.XQBridge -i(nsert) <common options> <file1> <file2> [filei]*");
        System.out.println();
        System.out.println("-q, -query:  query mode");
        System.out.println("file1:       XQuery file");
        System.out.println(
            "file2:       optional output file, output to stdout by default");
        System.out.println();
        System.out.println("-i, -insert: mapping mode");
        System.out.println("file1:       mapping file");
        System.out.println(
            "file2:       XML document to be inserted according to the mapping");
        System.out.println(
            "filei:       optional additional XML documents to be inserted according to the mapping");
        System.out.println();
        System.out.println("Common options:");
        System.out.println("-?, -help:       this message");
        System.out.println(
            "-conf:           XQBridge configuration file (see manual for details).");
        System.out.println(
            "-p, properties:  properties file containing database connection information.");
        System.out.println("-db:             JDBC database URL.");
        System.out.println("-user:           database user name.");
        System.out.println("-password:       database user password.");
		System.out.println("-cp, -classpath: classpath used to load JDBC drivers and user generators.");
        System.out.println(
            "-f, file:        file containing a list of files to be processed.");
        System.out.println(
            "                 Each line of the file has the number of files described above, according to the mode.");
        System.out.println(
            "                 Lines starting with # are comments");
    }

}
