/*
 * This file belongs to the XQuark distribution.
 * Copyright (C) 2004 XQuark Group.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquark.org.
 */

package org.xquark.xquery.parser;

import java.util.*;

import org.xquark.xpath.Axis;
import org.xquark.xpath.NodeKind;
import org.xquark.xquery.parser.util.Constants;

public class PrintVisitor extends DefaultParserVisitor {

	private StringBuffer buf = new StringBuffer();
	private PrintVisitor pvisitor = null;
	private int indent = 0;
	private boolean linefeed = false;
	private boolean imports = false;

	PrintVisitor() {
	}

	public void reset(boolean linefeed) {
		this.linefeed = linefeed;
		buf.setLength(0);
	}
	public void reset(boolean linefeed, boolean imports) {
		this.linefeed = linefeed;
		this.imports = imports;
		buf.setLength(0);
	}
	public String getString() {
		return buf.toString();
	}

	private void lineFeed() {
		if (linefeed) {
			buf.append('\n');
			for (int i=0;i<indent;i++) {
				buf.append('\t');
			}
		}
	}
	
	private String foreSee(XQueryExpression arg) throws XQueryException {
		if (pvisitor == null)
			pvisitor = new PrintVisitor();
		pvisitor.reset(linefeed);
		arg.accept(pvisitor);
		return pvisitor.getString();
	}

	//	public void visit(AggregateFunctionCall arg) throws XQueryException;
	public void visit(AttributeValuePair arg) throws XQueryException {
		if (arg.isComputed()) {
			buf.append("attribute ");
			if (!(arg.getExpression1() instanceof QName))
				buf.append("{ ");
			arg.getExpression1().accept(this);
			if (!(arg.getExpression1() instanceof QName))
				buf.append(" }");
			buf.append(" { ");
			arg.getExpression2().accept(this);
			buf.append(" }");
		} else {
			arg.getExpression1().accept(this);
			String expr2 = foreSee(arg.getExpression2());
			if (!(expr2.charAt(0) == '\"')) {
				buf.append("=\"");
				if (!(arg.getExpression2() instanceof Value))
					buf.append("{");
				buf.append(expr2);
				if (!(arg.getExpression2() instanceof Value))
					buf.append("}");
				buf.append("\"");
			} else {
				buf.append("=");
				if (!(arg.getExpression2() instanceof Value))
					buf.append("{");
				buf.append(expr2);
				if (!(arg.getExpression2() instanceof Value))
					buf.append("}");
			}
		}
	}
	public void visit(BinOpANDExpression arg) throws XQueryException {
		buf.append("(");
		arg.getExpression1().accept(this);
		buf.append(Constants.NODELISTOPERATORSSTRINGS[Constants.AND_NODELISTOPERATOR]);
		arg.getExpression2().accept(this);
		buf.append(")");
	}
	public void visit(BinOpORExpression arg) throws XQueryException {
		arg.getExpression1().accept(this);
		buf.append(Constants.NODELISTOPERATORSSTRINGS[Constants.OR_NODELISTOPERATOR]);
		arg.getExpression2().accept(this);
	}
	public void visit(CaseClauseExpression arg) throws XQueryException {
		buf.append("case ");
		if (arg.getSequenceType() != null) 
			arg.getSequenceType().accept(this);
		if (arg.getVariable() != null) 
			arg.getVariable().accept(this);;
		buf.append(" return ");
		arg.getReturnExpression().accept(this);
		
	}
	public void visit(CastTreatExpression arg) throws XQueryException {
		
		buf.append(Constants.TYPINGSTRINGS[arg.getOperator()]);
		buf.append(" as ");
		arg.getSequenceType().accept(this);
		buf.append("(");
		arg.getExpression().accept(this);
		buf.append(")");
		
	}
	public void visit(CData arg) throws XQueryException {
		
		buf.append("<![CDATA[ ");
		buf.append(arg.getCData());
		buf.append(" ]]>");
		
	}
	public void visit(ComputedNamespace arg) throws XQueryException {
		
		buf.append("namespace ");
		buf.append(arg.getNcname());
		buf.append(" { ");
		arg.getExpression().accept(this);
		buf.append(" }");
		
	}
	public void visit(ComputedText arg) throws XQueryException {
		
		buf.append("text {");
		arg.getExpression().accept(this);
		buf.append(" }");
		
	}
	
	public void visit(Document arg) throws XQueryException {
		
		buf.append("document { ");
		arg.getExpression().accept(this);;
		buf.append(" }");
		
	}

	public void visit(Element arg) throws XQueryException {
		
		if (arg.isComputed()) {
			buf.append("element ");
			if (!(arg.getStartTag() instanceof QName))
				buf.append("{ ");
			arg.getStartTag().accept(this);
			if (!(arg.getStartTag() instanceof QName))
				buf.append(" }");
			buf.append(" {");
			ArrayList attributes = arg.getAttributes();
			if (attributes != null)
				for (int i = 0; i < attributes.size(); i++) {
					buf.append(" ");
					((AttributeValuePair) attributes.get(i)).accept(this);
					if (i < attributes.size() - 1)
						buf.append(",");
				}
			ArrayList subExpressions = arg.getSubExpressions();
			if (subExpressions != null)
				for (int j = 0; j < subExpressions.size(); j++) {
					if (j == 0 && attributes != null)
						buf.append(",");
					buf.append(" ");
					((XQueryExpression) subExpressions.get(j)).accept(this);
					if (j < subExpressions.size() - 1)
						buf.append(",");
				}
			buf.append(" }");
		} else {
			buf.append("<");
			arg.getStartTag().accept(this);
			ArrayList attributes = arg.getAttributes();
			if (attributes != null)
				for (int i = 0; i < attributes.size(); i++) {
					buf.append(" ");
					((AttributeValuePair) attributes.get(i)).accept(this);
				}
			buf.append(">");
			ArrayList subExpressions = arg.getSubExpressions();
			if (subExpressions != null) {
				indent++;
				for (int j = 0; j < subExpressions.size(); j++) {
					lineFeed();
					XQueryExpression exprj = (XQueryExpression) subExpressions.get(j);
					if (!(exprj instanceof ValueText) && !( exprj instanceof Element))
						buf.append("{");
					((XQueryExpression) subExpressions.get(j)).accept(this);
					if (!(exprj instanceof ValueText) && !( exprj instanceof Element))
						buf.append("}");
				}
				indent--;
				lineFeed();
			}
			buf.append("</");
			arg.getStartTag().accept(this);
			buf.append(">");
		}
		
	}
	public void visit(ElementTest arg) throws XQueryException {
		
		buf.append("element(");
		if (arg.getSchemaContextPath() != null) 
			arg.getSchemaContextPath().accept(this);
		else if (arg.getDeclQName() != null) {
			arg.getDeclQName().accept(this);
			if (arg.getTypeQName() != null) {
				buf.append(",");
				arg.getTypeQName().accept(this);
				if (arg.getNillable())
					buf.append(" nillable");			
			}
		}
		buf.append(')');
		
	}

	//public void visit(ExternalVariable arg) throws XQueryException {
	
	public void visit(FLWRExpression arg) throws XQueryException {
		if (arg.getParentExpression() != null && !arg.getParentExpression().isEmpty()) {
			indent++;
			lineFeed();
		}		
		
		ArrayList variables = arg.getVariables();
		if (variables != null) {
			for (int i = 0; i < variables.size(); i++) {
				Variable vari = (Variable)variables.get(i);
				int bindingtype = vari.getBindingType();
				switch(bindingtype) {
					case Constants.FOR_BINDINGTYPE :
					buf.append(" for ");
					vari.accept(this);
					buf.append(" in ");
					break;
				case Constants.LET_BINDINGTYPE :
					buf.append(" let ");
					vari.accept(this);
					buf.append(" := ");
					break;
				case Constants.OUTERFOR_BINDINGTYPE :
					buf.append(" outerfor ");
					vari.accept(this);
					buf.append(" in ");
					break;
				case Constants.OUTERLET_BINDINGTYPE :
					buf.append(" outerlet ");
					vari.accept(this);
					buf.append(" := ");
				}
				if (vari.getPositionVar() != null) {
					buf.append(" at ");
					vari.getPositionVar().accept(this);
				}
				if (vari.getExpression() != null)
					buf.append(vari.getExpression());
			}
		}
		if (arg.getWhereClause() != null) {
			lineFeed();
			buf.append(" where ");
			arg.getWhereClause().accept(this);
		}
		if (arg.getOrderBy() != null) {
			lineFeed();
			buf.append(" order by ");
			for (int i=0, j=0; i < arg.getOrderBy().size(); i++) {
				XQueryExpression expri = (XQueryExpression) arg.getOrderBy().get(i);
				if (expri.getOrder(0) == Constants.PLACE_ORDER) {
					j++;
					continue;
				} 
				if (i != j)
					buf.append(",");
				expri.accept(this);
				if (expri.getOrder(0) != Constants.NOTHING) {
					buf.append(" ");
					buf.append(Constants.ORDERSSTRINGS[expri.getOrder(0)]);
				}
				if (expri.getOrder(1) != Constants.NOTHING) {
					buf.append(" empty ");
					buf.append(Constants.ORDERSSTRINGS[expri.getOrder(1)]);
				}
			}
		}
		if (arg.getReturnClause() != null) {
			lineFeed();
			buf.append(" return ");
			indent++;
			lineFeed();
			arg.getReturnClause().accept(this);
			indent--;
			lineFeed();
		}
		
		if (arg.getParentExpression() != null && !arg.getParentExpression().isEmpty()) {
			indent--;
		}		
	}
	public void visit(FunctionCall arg) throws XQueryException {
		
		arg.getFuncName().accept(this);
		buf.append("(");
		ArrayList arguments = arg.getArguments();
		if (arguments != null) {
			for (int i = 0; i < arguments.size(); i++) {
				if (i != 0)
					buf.append(",");
				((XQueryExpression) arguments.get(i)).accept(this);
			}
		}
		buf.append(")");
		
	}
	public void visit(FunctionDeclaration arg) throws XQueryException {
		
		buf.append("declare function ");
		arg.getFuncName().accept(this);
		buf.append(" (");
		if (arg.getParameters() != null) {
			for (int i = 0; i < arg.getParameters().size(); i++) {
				if (i != 0)
					buf.append(",");
				Variable vari = (Variable)arg.getParameters().get(i);
				vari.accept(this);
				if (vari.getTypeDeclaration() != null) {
					buf.append(" as ");
					vari.getTypeDeclaration().accept(this);
				}
			}
		}
		buf.append(")");
		if (arg.getReturnType() != null) {
			buf.append(" as ");
			arg.getReturnType().accept(this);
		}
		if (arg.getExpressions() != null) {
			buf.append(" { ");
			for (int i = 0; i < arg.getExpressions().size(); i++) {
				if (i != 0)
					buf.append(" , ");
				((XQueryExpression) arg.getExpressions().get(i)).accept(this);
			}
			buf.append(" } ");
		} else {
			buf.append(" external");
		}
		
	}
	public void visit(ImportModule arg) throws XQueryException {
		
		buf.append("import module ");
		if (arg.getPrefix() != null) {
			buf.append("namespace ");
			buf.append(arg.getPrefix());
			buf.append(" = ");
		}
		buf.append("\"");
		buf.append(arg.getName());
		buf.append("\"");
		if (arg.getLocation() != null) {
			buf.append(" at \"");
			buf.append(arg.getLocation());
			buf.append("\"");
		}
		buf.append(";");
		
	}
	public void visit(ImportSchema arg) throws XQueryException {
		
		buf.append("import schema ");
		if (arg.getHasPrefix()) {
			if (arg.getPrefix() == null)
				buf.append("default element namespace ");
			else {
				buf.append("namespace ");
				buf.append(arg.getPrefix());
				buf.append(" = ");
			}
		}
		buf.append("\"");
		buf.append(arg.getName());
		buf.append("\"");
		if (arg.getLocation() != null) {
			buf.append(" at \"");
			buf.append(arg.getLocation());
			buf.append("\"");
		}
		buf.append(";");
		
	}
	//  public void visit(InputFunctionCall arg) throws XQueryException {
	public void visit(InstanceOfExpression arg) throws XQueryException {
		
		arg.getExpression().accept(this);
		buf.append(" instanceof ");
		arg.getSequenceType().accept(this);
		
	}
	//	public void visit(InternalFunctionCall arg) throws XQueryException;
	public void visit(ITEExpression arg) throws XQueryException {
		
		buf.append(" if ( ");
		arg.getIfExpression().accept(this);
		buf.append(" ) then ");
		arg.getThenExpression().accept(this);
		buf.append(" else ");
		arg.getElseExpression().accept(this);
		
	}
	public void visit(ItemType arg) throws XQueryException {
		
		if (arg.getExpression() != null)
			arg.getExpression().accept(this);
		else
			buf.append("item()");
		
	}
	//	public void visit(LibraryFunctionCall arg) throws XQueryException;
	public void visit(ListOpAFTERExpression arg) throws XQueryException {
		
		arg.getExpression1().accept(this);
		buf.append(" after ");
		arg.getExpression2().accept(this);
		
	}
	public void visit(ListOpArithExpression arg) throws XQueryException {
		
		arg.getExpression1().accept(this);
		buf.append(" ");
		buf.append(Constants.ARITHMETICSSTRINGS[arg.getOperator()]);
		buf.append(" ");
		arg.getExpression2().accept(this);
		
	}
	public void visit(ListOpBEFOREExpression arg) throws XQueryException {
		
		arg.getExpression1().accept(this);
		buf.append(" before ");
		arg.getExpression2().accept(this);
		
	}
	public void visit(ListOpCompExpression arg) throws XQueryException {
		
		arg.getExpression1().accept(this);
		buf.append(" ");
		buf.append(Constants.COMPOPSSTRINGS[arg.getOperator()]);
		buf.append(" ");
		arg.getExpression2().accept(this);
		
	}
	public void visit(ListOpEXCEPTExpression arg) throws XQueryException {
		
		arg.getExpression1().accept(this);
		buf.append(" except ");
		arg.getExpression2().accept(this);
		
	}
	public void visit(ListOpINTERSECTExpression arg) throws XQueryException {
		
		arg.getExpression1().accept(this);
		buf.append(" intersect ");
		arg.getExpression2().accept(this);
		
	}
	public void visit(ListOpUNIONExpression arg) throws XQueryException {
		
		arg.getExpression1().accept(this);
		buf.append(" union ");
		arg.getExpression2().accept(this);
		
	}
	public void visit(LocatedExpression arg) throws XQueryException {
		
		for (int i = 0; i < arg.getSteps().size(); i++)
			((Step)arg.getSteps().get(i)).accept(this);
		
	}
	public void visit(NodeTest arg) throws XQueryException {
		
		buf.append(NodeKind.NODEKINDSTRINGS[arg.getKind()]);
		buf.append('(');
		if (arg.getArg() != null) 
			arg.getArg().accept(this);
		buf.append(')');
		
	}
	//  public void visit(PrimitiveFunctionCall arg) throws XQueryException;
	public void visit(QName arg) throws XQueryException {
		
		if (arg.getPrefixName() != null && (!arg.getPrefixName().equals("*") || !arg.getLocalName().equals("*"))) {
			buf.append(arg.getPrefixName());
			buf.append(":");
		}
		buf.append(arg.getLocalName());
		
	}
	public void visit(QuantifiedExpression arg) throws XQueryException {
		
		buf.append(Constants.QUANTIFIERSSTRINGS[arg.getKind()]);
		buf.append(" ");
		for (int i = 0; i < arg.getVariables().size(); i++) {
			if (i != 0)
				buf.append(" , ");
			Variable vari = (Variable)arg.getVariables().get(i);				
			vari.accept(this);
			buf.append(" in ");
			buf.append(vari.getExpression());
		}
		buf.append(" satisfies ");
		arg.getExpression().accept(this);
		
	}
	public void visit(RangeExpression arg) throws XQueryException {
		
		arg.getExpression1().accept(this);
		buf.append(" to ");
		arg.getExpression2().accept(this);
		
	}
	public void visit(SchemaContextPath arg) throws XQueryException {
		
		for (int i = 0; i < arg.getSteps().size(); i++)
			((SchemaContextStep) arg.getSteps().get(i)).accept(this);
		
	}
	public void visit(SchemaContextStep arg) throws XQueryException {
		
		if (arg.getAxis() == Constants.WITHAXIS)
			buf.append('/');
		else if (arg.getAxis() == Constants.WITHTYPE)
			buf.append("type(");
		arg.getExpression().accept(this);
		if (arg.getAxis() == Constants.WITHTYPE)
			buf.append(')');
		
	}
	public void visit(SequenceType arg) throws XQueryException {
		
		if (arg.getItemType() != null) {
			arg.getItemType().accept(this);
			if (arg.getOccurrence() != -1)
				buf.append(Constants.OCCURRENCE_INDICATORS[arg.getOccurrence()]);
		} else
			buf.append("empty()");
		
	}
	public void visit(SortedExpression arg) throws XQueryException {
		
		arg.getExpression().accept(this);;
		if (arg.getStable())
			buf.append(" stable");
		buf.append(" sortby ( ");
		for (int j = 0; j < arg.getSortClauses().size(); j++) {
			XQueryExpression expri = (XQueryExpression) arg.getSortClauses().get(j);
			if (j != 0)
				buf.append(",");
			expri.accept(this);
			if (expri.getOrder(0) != Constants.NOTHING) {
				buf.append(" ");
				buf.append(Constants.ORDERSSTRINGS[expri.getOrder(0)]);
			}
			if (expri.getOrder(1) != Constants.NOTHING) {
				buf.append(" empty ");
				buf.append(Constants.ORDERSSTRINGS[expri.getOrder(1)]);
			}
		}
		buf.append(" )");
		
	}
	public void visit(Step arg) throws XQueryException {
		
		if (arg.hasSeparator())
			buf.append('/');
		buf.append(Axis.AXISSTRINGS[arg.getAxis()]);
		arg.getExpression().accept(this);
		if (arg.getPredicates() != null)
			for (int i = 0; i < arg.getPredicates().size(); i++) {
				buf.append("[");
				((XQueryExpression) arg.getPredicates().get(i)).accept(this);
				buf.append("]");
			}
		
	}
	public void visit(TypeSwitchExpression arg) throws XQueryException {
		
		buf.append("typeswitch (");
		arg.getExpression1().accept(this);
		buf.append(") ");
		for (Iterator it = arg.getCases().values().iterator();it.hasNext();) {
			((CaseClauseExpression)it.next()).accept(this);			
		}
		buf.append(" default ");
		if (arg.getDefaultVariable() != null)
			arg.getDefaultVariable().accept(this);
		arg.getExpression2().accept(this);
		
	}
	public void visit(UnOpMinusExpression arg) throws XQueryException {
		
		if (arg.getMinus()) 
			buf.append("- ");
		else 
			buf.append("+ ");
		arg.getExpression().accept(this);
		
	}
	public void visit(ValidateExpression arg) throws XQueryException {
		
		buf.append("validate");
		if (arg.getValidationType() != Constants.NOTHING) {
			buf.append(" ");
			buf.append(Constants.VALIDATIONSTRINGS[arg.getValidationType()]);
		}
		if (arg.getScopeType() != Constants.NOTHING) {
			buf.append(" ");
			buf.append(Constants.SCOPE_STRINGS[arg.getScopeType()]);
		}		
		if (arg.getSchemaContext() != null) {
			buf.append(" ");
			arg.getSchemaContext().accept(this);
		}
		buf.append(" { ");
		arg.getExpression().accept(this);
		buf.append(" }");
		
	}
	public void visit(Value arg) throws XQueryException {
		
		buf.append(arg.getValue());
		
	}
	//	public void visit(ValueBoolean arg) throws XQueryException;
	//	public void visit(ValueDecimal arg) throws XQueryException;
	//	public void visit(ValueDouble arg) throws XQueryException;
	//	public void visit(ValueFloat arg) throws XQueryException;
	//	public void visit(ValueInteger arg) throws XQueryException;
	public void visit(ValueString arg) throws XQueryException {
		
		buf.append("\"");
		buf.append(arg.getValue());
		buf.append("\"");
				
	}
	//	public void visit(ValueParameter arg) throws XQueryException;
	//	public void visit(ValueText arg) throws XQueryException;
	public void visit(Variable arg) throws XQueryException {
		
		buf.append("$");
		arg.getVarName().accept(this);
		
	}
	public void visit(XMLComment arg) throws XQueryException {
		
		if (arg.isComputed()) {
			buf.append("comment { ");
			arg.getExpression().accept(this);
			buf.append(" }");
		} else {
			buf.append("<!-- ");
			arg.getExpression().accept(this);
			buf.append(" -->");
		}
		
	}
	public void visit(XMLProcessingInstruction arg) throws XQueryException {
		
		if (arg.isComputed()) {
			buf.append("processing-instruction ");
			arg.getExpression1().accept(this);
			buf.append(" { ");
			arg.getExpression2().accept(this);
			buf.append(" }");
		} else {
			StringBuffer buf = new StringBuffer();
			buf.append("<? ");
			arg.getExpression1().accept(this);
			buf.append(" ");
			arg.getExpression2().accept(this);
			buf.append(" ?>");
		}
		
	}
	//	public void visit(XQueryBinaryOperatorExpression arg) throws XQueryException;
	//	public void visit(XQueryBooleanOperatorExpression arg) throws XQueryException;
	//  public void visit(XQueryExpression arg) throws XQueryException {
	public void visit(XQueryExpressionSequence arg) throws XQueryException {
		
		for (int j=0;j<arg.getSubExpressions().size();j++) {
			if (j!=0) buf.append(" , ");
			((XQueryExpression)arg.getSubExpressions().get(j)).accept(this);
		}
		
	}
	//	public void visit(XQueryListBinaryOperatorExpression arg) throws XQueryException;
	//	public void visit(XQueryListUnaryOperatorExpression arg) throws XQueryException;
	//	public void visit(XQueryUnaryOperatorExpression arg) throws XQueryException;
	public void visit(XQueryModule arg) throws XQueryException {
		if (arg.getVersion() != null) {
			lineFeed();
			buf.append("xquery version \"");
			buf.append(arg.getVersion());
			buf.append("\";");
		}
		if (arg.getPrefix() != null && arg.getNamespace() != null) {
			lineFeed();
			buf.append("module namespace ");
			buf.append(arg.getPrefix());
			buf.append(" = \"");
			buf.append(arg.getNamespace());
			buf.append("\"");
		}
		if (arg.getDeclarations() != null) {
			lineFeed();
			Collection declarationsPrefixes = arg.getDeclarations().getPrefixes();
			Iterator prefixes = declarationsPrefixes.iterator();
			while (prefixes.hasNext()) {
				String prefix = (String) prefixes.next();
				if ((prefix == null) || (prefix.trim().length() == 0)) {
					buf.append("declare default element namespace ");
					buf.append("=\"");
					buf.append((String) arg.getDeclarations().getNamespaceURI(prefix));
					buf.append("\";");
				} else if (prefix != "xml") {
					buf.append("declare namespace ");
					buf.append(prefix);
					buf.append("=\"");
					buf.append((String) arg.getDeclarations().getNamespaceURI(prefix));
					buf.append("\";");
				}
			}
		}
		if (arg.getXmlSpace() != Constants.NOTHING) {
			lineFeed();
			buf.append("declare xmlspace ");
			buf.append(Constants.XMLSPACESTRINGS[arg.getXmlSpace()]);
		}
		if (arg.getDefaultFunctionNamespace() != null) {
			lineFeed();
			buf.append("declare default function namespace = \"");
			buf.append(arg.getDefaultFunctionNamespace());
			buf.append("\";");
		}
		if (arg.getDefaultCollation() != null) {
			lineFeed();
			buf.append("declare default collation = \"");
			buf.append(arg.getDefaultCollation());
			buf.append("\";");
		}
		if (arg.getBaseURI() != null) {
			lineFeed();
			buf.append("declare base-uri = \"");
			buf.append(arg.getBaseURI());
			buf.append("\";");
		}
		if (arg.getSchemaList() != null) {
			for (int i=0;i<arg.getSchemaList().size();i++) {
				lineFeed();
				((ImportSchema) arg.getImportSchemas().get(arg.getSchemaList().get(i))).accept(this);
			}
		}
		if (imports) {
			if (arg.getModuleList() != null) {
				for (int i=0;i<arg.getModuleList().size();i++) {
					lineFeed();
					((ImportModule) arg.getImportModules().get(arg.getModuleList().get(i))).accept(this);
				}
			}
			if (arg.getVariableList() != null) {
				for (int i=0;i<arg.getVariableList().size();i++) {
					lineFeed();
					QName name = (QName)arg.getVariableList().get(i);
					HashMap map = (HashMap)arg.getVariables().get(name.getNameSpace());
					Variable vari = (Variable) map.get(foreSee(name));
					buf.append("declare variable ");
					vari.accept(this);
					if (vari.getExpression() != null) {
						buf.append(" ");
						buf.append(vari.getExpression());
					}
					buf.append(" external");
				}
			}
		}
		if (arg.getValidation() != Constants.NOTHING && arg.getValidation() != Constants.VALIDATION_PRESERVE) {
			lineFeed();
			buf.append("declare validation ");
			buf.append(Constants.VALIDATIONSTRINGS[arg.getValidation()]);
		}
		if (arg.getFunctionList() != null) {
			for (int i=0;i<arg.getFunctionList().size();i++) {
				lineFeed();
				QName name = (QName)arg.getFunctionList().get(i);
				HashMap map = (HashMap)arg.getFunctions().get(name.getNameSpace());
				((FunctionDeclaration) map.get(name.getLocalName())).accept(this);
			}
		}
		if (arg.getExpressions() != null) {
			for (int i = 0; i < arg.getExpressions().size(); i++) {
				if (i != 0) {
					buf.append(" ,");
					lineFeed();
				}
				((XQueryExpression) arg.getExpressions().get(i)).accept(this);
			}
		}
	}
	//public void visit(XQueryVoid arg) throws XQueryException {
	//	
	//	// primitive functions
	//	
	//	public void visit(FunctionABS arg) throws XQueryException ;
	//	public void visit(FunctionAVG arg) throws XQueryException ;
	//	public void visit(FunctionCEILING arg) throws XQueryException;
	//	public void visit(FunctionCOLLECTION arg) throws XQueryException;
	//	public void visit(FunctionCONCAT arg) throws XQueryException;
	//	public void visit(FunctionCONTAINS arg) throws XQueryException;
	//	public void visit(FunctionCOUNT arg) throws XQueryException;
	//	public void visit(FunctionCURRENT_DATE arg) throws XQueryException;
	//	public void visit(FunctionCURRENT_DATETIME arg) throws XQueryException;
	//	public void visit(FunctionCURRENT_TIME arg) throws XQueryException;
	//	public void visit(FunctionDATA arg) throws XQueryException;
	//	public void visit(FunctionDATE arg) throws XQueryException;
	//	public void visit(FunctionDATETIME arg) throws XQueryException;
	//	public void visit(FunctionDECIMAL arg) throws XQueryException;
	//	public void visit(FunctionDISTINCT_VALUES arg) throws XQueryException;
	//	public void visit(FunctionDOC arg) throws XQueryException;
	//	public void visit(FunctionDOUBLE arg) throws XQueryException;
	//	public void visit(FunctionEMPTY arg) throws XQueryException;
	//	public void visit(FunctionENDS_WITH arg) throws XQueryException;
	//	public void visit(FunctionEXISTS arg) throws XQueryException;
	//	public void visit(FunctionGET_MONTH_FROM_DATE arg) throws XQueryException;
	//	public void visit(FunctionGET_YEAR_FROM_DATE arg) throws XQueryException;
	//	public void visit(FunctionFALSE arg) throws XQueryException;
	//	public void visit(FunctionFLOOR arg) throws XQueryException;
	//	public void visit(FunctionINTEGER arg) throws XQueryException;
	//	public void visit(FunctionLOCAL_NAME arg) throws XQueryException;
	//	public void visit(FunctionLOWER_CASE arg) throws XQueryException;
	//	public void visit(FunctionMATCHES arg) throws XQueryException;
	//	public void visit(FunctionMAX arg) throws XQueryException;
	//	public void visit(FunctionMIN arg) throws XQueryException;
	//	public void visit(FunctionNAMESPACE_URI arg) throws XQueryException;
	//	public void visit(FunctionNOT arg) throws XQueryException;
	//	public void visit(FunctionNUMBER arg) throws XQueryException;
	//	public void visit(FunctionROUND arg) throws XQueryException;
	//	public void visit(FunctionSTARTS_WITH arg) throws XQueryException;
	//	public void visit(FunctionSTRING arg) throws XQueryException;
	//	public void visit(FunctionSTRING_LENGTH arg) throws XQueryException;
	//	public void visit(FunctionSUBSTRING arg) throws XQueryException;
	//	public void visit(FunctionSUM arg) throws XQueryException;
	//	public void visit(FunctionTIME arg) throws XQueryException;
	//	public void visit(FunctionTRUE arg) throws XQueryException;
	//	public void visit(FunctionUPPER_CASE arg) throws XQueryException;
	//	public void visit(FunctionPART arg) throws XQueryException;
	//	public void visit(FunctionTUPLE arg) throws XQueryException;
	//	public void visit(FunctionDEEP_EQUALS arg) throws XQueryException;
}
